#!/usr/bin/perl

package base_ops;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(println startSmt finishSmt declareVar bvconst assert lnot land lor bvnot bvand bvor bvxor equal slice log2 ceil_log2 next_2n);

use POSIX qw/ceil/;


use constant {
	UNSIGNED => "u",
	SIGNED => "s"
};


$varCounter = 0;

sub declareVar {
	@_ == 2 || die("ERROR: too few/many arguments for declareVar()\n\tusage: declareVar(varName,bitWidth)\n");

	local $var = $_[0]."_$varCounter";
	local $bw = $_[1];
	
	println("(declare-fun $var () (_ BitVec $bw))");
	
	$varCounter++;
	
	return $var;
}


sub bvconst {
	@_ == 2 || die("ERROR: too few/many arguments for bvconst()\n\tusage: bvconst(const,bitWidth)\n");

	local $const = $_[0];
	local $bw = $_[1];

	return "(_ bv$const $bw)";
}


sub assert {
	@_ == 1 || die("ERROR: too few/many arguments for assert()\n\tusage: assert(cond)\n");

	local $cond = $_[0];

	println("(assert $cond)");
}


sub lnot {
	@_ == 1 || die("ERROR: too few/many arguments for lnot()\n\tusage: lnot(cond)\n");

	local $cond = $_[0];

	return "(not $cond)";
}

sub land {
	@_ == 2 || die("ERROR: too few/many arguments for land()\n\tusage: land(cond1,cond2)\n");

	local $cond1 = $_[0];
	local $cond2 = $_[1];

	return "(and $cond1 $cond2)";
}

sub lor {
	@_ == 2 || die("ERROR: too few/many arguments for lor()\n\tusage: lor(cond1,cond2)\n");

	local $cond1 = $_[0];
	local $cond2 = $_[1];

	return "(or $cond1 $cond2)";
}


sub bvnot {
	@_ == 1 || die("ERROR: too few/many arguments for bvnot()\n\tusage: bvnot(term)\n");

	local $term = $_[0];

	return "(bvnot $term)";
}

sub bvand {
	@_ == 2 || die("ERROR: too few/many arguments for bvand()\n\tusage: bvand(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvand $term1 $term2)";
}

sub bvor {
	@_ == 2 || die("ERROR: too few/many arguments for bvor()\n\tusage: bvor(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvor $term1 $term2)";
}

sub bvxor {
	@_ == 2 || die("ERROR: too few/many arguments for bvxor()\n\tusage: bvxor(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvxor $term1 $term2)";
}


sub equal {
	@_ == 2 || die("ERROR: too few/many arguments for equal()\n\tusage: equal(input1,input2)\n");

	local $input1 = $_[0];
	local $input2 = $_[1];

	return "(= $input1 $input2)";
}

sub slice {
	@_ == 2 || @_ == 3 || die("ERROR: too few/many arguments for slice()\n\tusage: slice(var,index[,index2])\n");

	local $var = $_[0];
	local $index = $_[1];
	local $index2 = (@_ == 3 ? $_[2] : $index);

	return "((_ extract $index $index2) $var)";
}



sub log2 {
	@_ == 1 || die("ERROR: too few/many arguments for log2()\n\tusage: log2(n)\n");

	local $n = $_[0];

	return log($n)/log(2);
}

sub ceil_log2 {
	@_ == 1 || die("ERROR: too few/many arguments for ceil_log2()\n\tusage: ceil_log2(n)\n");

	local $n = $_[0];

	return ceil(log2($n));
}

sub next_2n {
	@_ == 1 || die("ERROR: too few/many arguments for ceil_2n()\n\tusage: ceil_2n(n)\n");

	local $n = $_[0];

	return 1<<ceil_log2($n);
}



sub println {
	local $,="";
	print +( @_ ? @_ : $_ ), "\n";
}


sub startSmt {
	@_ == 1 || @_ == 2 || die("ERROR: too few/many arguments for startSmt()\n\tusage: startSmt(operation[,ops_translate_to])\n");
	
	local $ops_translate_to = (@_ == 1)? "base operations (bitwise ones, equality, and slicing)" : $_[1];

	$varCounter = 0;

	print <<_EOS_;
(set-logic QF_BV)
(set-info :source |
Checking the soundness of the translation of $_[0] to $ops_translate_to.
Contributed by Gergely Kovasznai, Andreas Froehlich, and Armin Biere.
Institute for Formal Models and Verification, JKU, Linz, 2013.
source: http://fmv.jku.at/smtbench
|)
(set-info :smt-lib-version 2.0)
(set-info :category "crafted")
(set-info :status unsat)
_EOS_
}

sub finishSmt {
	print <<_EOS_;
(check-sat)
(exit)
_EOS_
}


1;
