#!/usr/bin/perl 

package binmagic_by_bw_shl1;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(benchmarks_binmagic_by_bw_shl1);

use base_ops;
use helper;

###################################################
# Binary magic numbers by bitwise operations and left shift by 1
###################################################

sub bvshl {
	@_ == 2 || die("ERROR: too few/many arguments for bvshl()\n\tusage: bvshl(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvshl $term1 $term2)";
}


sub binmagic_by_bw_shl1 {
	@_ == 2 || die("ERROR: too few/many arguments for binmagic_by_bw_shl1()\n\tusage: binmagic_by_bw_shl1(bw1, bw2)\n");

	local $bw1 = $_[0];
	local $bw2 = $_[1];
	
	local $binmagic = ();
	
	$binmagic = declareVar("binmagic_0", $bw2);
	
	assert(
		equal(
			binmagic_by_bw_shl1::bvshl(
				$binmagic,
				bvconst(1, $bw2)
			),
			bvnot($binmagic)
	));
	
	local $cond = $binmagic;
	local $i;
	local $log_bw = ceil_log2($bw1);
	for ($i = 1; $i <= $log_bw; $i++) {
		$binmagic = declareVar("binmagic_$i", $bw2);
		
		assert(
			equal(
				binmagic_by_bw_shl1::bvshl(
					$binmagic,
					bvconst(1, $bw2)
				),
				bvxor(
					$binmagic,
					$cond
		)));
		
		$cond = bvand($cond, $binmagic);
	}
	
	return $binmagic;
}


sub test_binmagic_by_bw_shl1 {
	@_ == 2 || die("ERROR: too few/many arguments for test_binmagic_by_bw_shl1()\n\tusage: test_binmagic_by_bw_shl1(bw1, bw2)\n");


	# START

	startSmt("binmagic", "bitwise ops, equality, and left shift by 1");


	### INPUT VARIABLES

	local $bw1 = $_[0];
	local $bw2 = $_[1];


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw2);
	assert(
		equal(
			$result,
			binmagic($bw1, $bw2)
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				binmagic_by_bw_shl1($bw1, $bw2)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_binmagic_by_bw_shl1 {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_binmagic_by_bw_shl1()\n\tusage: benchmarks_binmagic_by_bw_shl1(dir, bw1_min, bw2_min, bw2_max)\n");
	local $dir = $_[0];
	local $bw1_min = $_[1];
	local $bw2_min = $_[2];
	local $bw2_max = $_[3];
	
	$bw1_min < $bw2_min || die("ERROR: bw1_min must be less then bw2_min in benchmarks_binmagic_by_bw_shl1\n");

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw1, $bw2;
	for ($bw2 = $bw2_min; $bw2 <= $bw2_max; $bw2++) {
	  local $_bw2 = 2**$bw2;
	  
		for ($bw1 = $bw1_min; $bw1 < $bw2; $bw1++) {
		  local $_bw1 = 2**$bw1;
	  
			println("Generating benchmark for binmagic by bitwise ops and left shift by 1 with bit-widths $_bw1 and $_bw2...");
		
			$filename = "$dir/binmagic_by_bw_shl1_".sprintf("%05d", $_bw2)."_".sprintf("%05d", $_bw1).".smt2";
			open FILE, ">", $filename or die $!;
		
			select FILE;
		
			test_binmagic_by_bw_shl1($_bw1, $_bw2);
		
			select STDOUT;
		
			close FILE;
		}
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: _binmagic_by_bw_shl1  dir  bw1_min  bw2_min  bw2_max\n";
		exit;
	}
	benchmarks_binmagic_by_bw_shl1($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
