#!/usr/bin/perl

package bvadd;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvadd benchmarks_bvadd);

use base_ops;
use bvshl;

###################################################
# Addition
###################################################


sub bvadd {
	@_ == 3 || @_ == 5 || die("ERROR: too few/many arguments for bvadd()\n\tusage: bvadd(term1,term2,bw[,overflow_type,overflow_result])\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $result = declareVar("bvadd_result", $bw);
	local $cin = declareVar("cin", $bw);
	local $cout = declareVar("cout", $bw);

	assert(
		equal(
			$result,
			bvxor(
				bvxor(
					$term1,
					$term2
				),
				$cin
	)));

	assert(
		equal(
			$cout,
			bvor(
				bvor(
					bvand($term1, $term2),
					bvand($term1, $cin)
				),
				bvand($term2, $cin)
	)));

	assert(
		equal(
			$cin,
			bvshl_const($cout,1,$bw)
#			$bw == 1 ? bvconst(0,1) : bvshl_const($cout,1,$bw)
	));

	# If overflow detection is on
	if (@_>3) {
		if ($_[3] == UNSIGNED) {
			$_[4] = slice($cout, $bw-1);
		}
	}

	return $result;
}


sub test_bvadd {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvadd()\n\tusage: test_bvadd(input1,input2,bw)\n");


	# START

	startSmt("bvadd");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvadd $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvadd($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvadd {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvadd()\n\tusage: benchmarks_bvadd(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(10))+1; $bw<=2000; $bw+=int(rand(100))) {
		println("Generating benchmark for bvadd with bit-width $bw...");
	
		$filename = "$dir/bvadd_".sprintf("%05d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvadd("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvadd  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvadd($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
