#!/usr/bin/perl

package bvashr;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvashr bvashr benchmarks_bvashr);

use base_ops;
use helper;
use bvlshr;

###################################################
# Arithmetic right shift
###################################################


sub bvashr {
	@_ == 3 || die("ERROR: too few/many arguments for bvashr()\n\tusage: bvashr(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	return ite(
		slice($term1, $bw-1),
		bvnot(
			bvlshr(
				bvnot($term1),
				$term2,
				$bw
			)
		),
		bvlshr(
			$term1,
			$term2,
			$bw
		),
		$bw
	);
}


sub test_bvashr {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvashr()\n\tusage: test_bvashr(input1,input2,bw)\n");


	# START

	startSmt("bvashr");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvashr $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvashr($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvashr {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvashr()\n\tusage: benchmarks_bvashr(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(5))+1; $bw<=2000; $bw+=int(rand(200))) {
		println("Generating benchmark for bvashr with bit-width $bw...");
		
		$filename = "$dir/bvashr_".sprintf("%04d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvashr("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvashr  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvashr($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
