#!/usr/bin/perl

package bvlshr;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvlshr_const bvlshr benchmarks_bvlshr);

use base_ops;
use concat;
use helper;
use bvult;


###################################################
# Logical right shift
###################################################


###################################################
# Logical right shift by constant

sub bvlshr_const {
	@_ == 3 || die("ERROR: too few/many arguments for bvlshr_const()\n\tusage: bvlshr_const(term,const,bw)\n");
  
	local $term = $_[0];
	local $const = $_[1];
	local $bw = $_[2];

	if ($const == 0) {
		return $term;
	}
	elsif ($const<$bw) {
		return concat(
			bvconst(0, $const), $const,
			slice($term, $bw-1, $const), $bw-$const
		);
	}
	else {
		return bvconst(0, $bw);
	}
}


###################################################
# Logical right shift by any term

sub bvlshr {
	@_ == 3 || die("ERROR: too few/many arguments for bvlshr()\n\tusage: bvlshr(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];
	
	local $x, $x_next;
	
	$x = declareVar("bvlshr_x0", $bw);
	assert(
		equal(
			$x,
			$term1
	));
	
	local $log_bw = ceil_log2($bw);
	local $i;
	for ($i=1; $i<=$log_bw; $i++) {
		$x_next = declareVar("bvlshr_x$i", $bw);
		assert(
			equal(
				$x_next,
				ite(
					slice($term2, $i-1),
					bvlshr_const($x, 1<<($i-1), $bw),
					$x,
					$bw
		)));
		
		$x = $x_next;
	}

	return ite(
		bvult(
			$term2,
			bvconst($bw, $bw),
			$bw
		),
		$x,
		bvconst(0, $bw),
		$bw
	);
}


sub test_bvlshr {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvlshr()\n\tusage: test_bvlshr(input1,input2,bw)\n");


	# START

	startSmt("bvlshr");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvlshr $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvlshr($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvlshr {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvlshr()\n\tusage: benchmarks_bvlshr(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(5))+1; $bw<=2000; $bw+=int(rand(200))) {
		println("Generating benchmark for bvlshr with bit-width $bw...");
		
		$filename = "$dir/bvlshr_".sprintf("%04d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvlshr("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvlshr  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvlshr($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
