#!/usr/bin/perl

package bvmul;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvmul benchmarks_bvmul);

use base_ops;
use helper;
use zero_extend;
use bvadd;
use bvlshr;

###################################################
# Multiplication
###################################################


sub bvmul {
	@_ == 3 || @_ == 5 || die("ERROR: too few/many arguments for bvmul()\n\tusage: bvmul(term1,term2,bw[,overflow_type,overflow_result])\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];
	
	local $x, $x_next, $b;

	local $n = next_2n($bw);
	
	$x = declareVar("bvmul_x0", $n**2);
	assert(
		equal(
			$x,
			bvand(
				selfconcat(
					zero_extend($term1, $bw, $n-$bw),
					$n,
					$n**2
				),
				expand(
					zero_extend($term2, $bw, $n-$bw),
					$n,
					$n**2
	))));
	
	local $log_bw = ceil_log2($bw);
	local $i;
	for ($i=1; $i<=$log_bw; $i++) {
		$b = declareVar("bvmul_b${\($i - 1)}", $n**2);
		assert(
			equal(
				$b,
				binmagic($n<<($i-1), $n**2)
		));
		
		$x_next = declareVar("bvmul_x$i", $n**2);
		assert(
			equal(
				$x_next,
				bvadd(
					bvand(
						$x,
						$b
					),
					bvlshr_const(
						bvand(
							$x,
							bvnot($b)
						),
						($n-1)<<($i-1),
						$n**2
					),
					$n**2
		)));
		
		$x = $x_next;
	}

	# If overflow detection is on
	if (@_>3) {
		if ($_[3] == UNSIGNED) {
			$_[4] = $bw == 1 ? "false" :
				lnot(
					equal(
						slice($x, 2*$bw-1, $bw),
						bvconst(0, $bw)
				));
		}
	}

	return slice($x, $bw-1, 0);
}


sub test_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvmul()\n\tusage: test_bvmul(input1,input2,bw)\n");


	# START

	startSmt("bvmul");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvmul $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvmul($input1, $input2, $bw)
	)));

	# FINISH

	finishSmt();
}


sub benchmarks_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for benchmarks_bvmul()\n\tusage: benchmarks_bvmul(dir, bw_min, bw_max)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw++) {
		println("Generating benchmark for bvmul with bit-width $bw...");
		
		$filename = "$dir/bvmul_".sprintf("%02d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvmul("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 2 ) {
		print "usage: bvmul  dir  bw_min  bw_max\n";
		exit;
	}
	benchmarks_bvmul($ARGV[0], $ARGV[1], $ARGV[2]);
}
