#!/usr/bin/perl

package bvneg;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvneg benchmarks_bvneg);

use base_ops;
use bvadd;


###################################################
# Unary minus
###################################################


sub bvneg {
	@_ == 2 || @_ == 3 || die("ERROR: too few/many arguments for bvneg()\n\tusage: bvneg(term,bw[,overflow_result])\n");
  
	local $term = $_[0];
	local $bw = $_[1];

	# If overflow detection is on
#	if (@_>2) {
#		$_[2] = slice($cout, $bw-1);
#	}

	return bvadd(
		bvnot($term),
		bvconst(1, $bw),
		$bw
	);
}


sub test_bvneg {
	@_ == 2 || die("ERROR: too few/many arguments for test_bvneg()\n\tusage: test_bvneg(input,bw)\n");


	# START

	startSmt("bvneg");


	### INPUT VARIABLES

	local $input = $_[0];
	local $bw = $_[1];

	local $input = declareVar($input, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvneg $input)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvneg($input, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvneg {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvneg()\n\tusage: benchmarks_bvneg(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(10))+1; $bw<=2000; $bw+=int(rand(100))) {
		println("Generating benchmark for bvneg with bit-width $bw...");
	
		$filename = "$dir/bvneg_".sprintf("%05d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvneg("input", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}

}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvneg  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvadd($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
