#!/usr/bin/perl

package bvsdiv;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvsdiv benchmarks_bvsdiv);

use base_ops;
use helper;
use bvneg;
use bvudiv;

###################################################
# Signed division
###################################################


sub bvsdiv {
	@_ == 3 || die("ERROR: too few/many arguments for bvsdiv()\n\tusage: bvsdiv(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $abs1 = declareVar("bvsdiv_abs1", $bw);
	local $abs2 = declareVar("bvsdiv_abs2", $bw);
	local $udiv = declareVar("bvsdiv_udiv", $bw);

	assert(
		equal(
			$abs1,
			ite(
				slice($term1, $bw-1),
				bvneg($term1, $bw),
				$term1,
				$bw
			)
	));

	assert(
		equal(
			$abs2,
			ite(
				slice($term2, $bw-1),
				bvneg($term2, $bw),
				$term2,
				$bw
			)
	));

	assert(
		equal(
			$udiv,
			bvudiv(
				$abs1,
				$abs2,
				$bw
			)
	));

	return ite(
		equal(
			slice($term1, $bw-1),
			slice($term2, $bw-1)
		),
		$udiv,
		bvneg($udiv, $bw),
		$bw
	);
}


sub test_bvsdiv {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvsdiv()\n\tusage: test_bvsdiv(input1,input2,bw)\n");


	# START

	startSmt("bvsdiv");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvsdiv $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvsdiv($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvsdiv {
	@_ == 3 || die("ERROR: too few/many arguments for benchmarks_bvsdiv()\n\tusage: benchmarks_bvsdiv(dir, bw_min, bw_max)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw++) {
		println("Generating benchmark for bvsdiv with bit-width $bw...");
	
		$filename = "$dir/bvsdiv_".sprintf("%02d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvsdiv("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 2 ) {
		print "usage: bvsdiv  dir  bw_min  bw_max\n";
		exit;
	}
	benchmarks_bvsdiv($ARGV[0], $ARGV[1], $ARGV[2]);
}
