#!/usr/bin/perl

package bvslt;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvslt benchmarks_bvslt);

use base_ops;
use bvult;


###################################################
# Signed less than
###################################################


sub bvslt {
	@_ == 3 || die("ERROR: too few/many arguments for bvslt()\n\tusage: bvslt(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];
	
	return lor(
		land(
			slice($term1, $bw-1),
			lnot(
				slice($term2, $bw-1)
			)
		),
		land(
			equal(
				slice($term1, $bw-1),
				slice($term2, $bw-1)
			),
			bvult($term1, $term2, $bw)
	));
}


sub test_bvslt {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvslt()\n\tusage: test_bvslt(input1,input2,bw)\n");


	# START

	startSmt("bvslt");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", 1);
	assert(
		equal(
			$result,
			"(bvslt $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvslt($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvslt {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvadd()\n\tusage: benchmarks_bvadd(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(10))+1; $bw<=2000; $bw+=int(rand(100))) {
		println("Generating benchmark for bvslt with bit-width $bw...");
	
		$filename = "$dir/bvslt_".sprintf("%04d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvslt("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvslt  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvslt($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
