#!/usr/bin/perl

package bvsmod;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvsmod benchmarks_bvsmod);

use base_ops;
use helper;
use bvadd;
use bvsrem;

###################################################
# Signed modulus
###################################################


sub bvsmod {
	@_ == 3 || die("ERROR: too few/many arguments for bvsmod()\n\tusage: bvsmod(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $srem = declareVar("bvsmod_srem", $bw);

	assert(
		equal(
			$srem,
			bvsrem(
				$term1,
				$term2,
				$bw
			)
	));

	return ite(
		lor(
			equal(
				$srem,
				bvconst(0, $bw)
			),
			equal(
				slice($term1, $bw-1),
				slice($term2, $bw-1)
			)
		),
		$srem,
		bvadd(
			$srem,
			$term2,
			$bw
		),
		$bw
	);
}


sub test_bvsmod {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvsmod()\n\tusage: test_bvsmod(input1,input2,bw)\n");


	# START

	startSmt("bvsmod");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvsmod $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvsmod($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvsmod {
	@_ == 3 || die("ERROR: too few/many arguments for benchmarks_bvsmod()\n\tusage: benchmarks_bvsmod(dir, bw_min, bw_max)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw++) {
		println("Generating benchmark for bvsmod with bit-width $bw...");
	
		$filename = "$dir/bvsmod_".sprintf("%02d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvsmod("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}

}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 2 ) {
		print "usage: bvsmod  dir  bw_min  bw_max\n";
		exit;
	}
	benchmarks_bvsmod($ARGV[0], $ARGV[1], $ARGV[2]);
}
