#!/usr/bin/perl

package bvsrem;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvsrem benchmarks_bvsrem);

use base_ops;
use helper;
use bvneg;
use bvurem;

###################################################
# Signed remainder
###################################################


sub bvsrem {
	@_ == 3 || die("ERROR: too few/many arguments for bvsrem()\n\tusage: bvsrem(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $abs1 = declareVar("bvsrem_abs1", $bw);
	local $abs2 = declareVar("bvsrem_abs2", $bw);
	local $urem = declareVar("bvsrem_urem", $bw);

	assert(
		equal(
			$abs1,
			ite(
				slice($term1, $bw-1),
				bvneg($term1, $bw),
				$term1,
				$bw
			)
	));

	assert(
		equal(
			$abs2,
			ite(
				slice($term2, $bw-1),
				bvneg($term2, $bw),
				$term2,
				$bw
			)
	));

	assert(
		equal(
			$urem,
			bvurem(
				$abs1,
				$abs2,
				$bw
			)
	));

	return ite(
		slice($term1, $bw-1),
		bvneg($urem, $bw),
		$urem,
		$bw
	);
}


sub test_bvsrem {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvsrem()\n\tusage: test_bvsrem(input1,input2,bw)\n");


	# START

	startSmt("bvsrem");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvsrem $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvsrem($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvsrem {
	@_ == 3 || die("ERROR: too few/many arguments for benchmarks_bvsrem()\n\tusage: benchmarks_bvsrem(dir, bw_min, bw_max)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw++) {
		println("Generating benchmark for bvsrem with bit-width $bw...");
	
		$filename = "$dir/bvsrem_".sprintf("%02d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvsrem("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 2 ) {
		print "usage: bvsrem  dir  bw_min  bw_max\n";
		exit;
	}
	benchmarks_bvsrem($ARGV[0], $ARGV[1], $ARGV[2]);
}
