#!/usr/bin/perl

package bvsub;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvsub benchmarks_bvsub);

use base_ops;
use bvneg;
use bvadd;


###################################################
# Subtraction
###################################################


sub bvsub {
	@_ == 3 || @_ == 5 || die("ERROR: too few/many arguments for bvneg()\n\tusage: bvneg(term1,term2,bw[,overflow_type,overflow_result])\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	# If overflow detection is on
#	if (@_>2) {
#		$_[2] = slice($cout, $bw-1);
#	}

	return bvadd(
		$term1,
		bvneg($term2, $bw),
		$bw
	);
}


sub test_bvsub {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvsub()\n\tusage: test_bvsub(input1,input2,bw)\n");


	# START

	startSmt("bvsub");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvsub $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvsub($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvsub {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvsub()\n\tusage: benchmarks_bvsub(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(50))+1; $bw<=2000; $bw+=int(rand(100))) {
		println("Generating benchmark for bvsub with bit-width $bw...");
	
		$filename = "$dir/bvsub_".sprintf("%05d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvsub("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvsub  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvsub($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
