#!/usr/bin/perl

package bvult;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvult benchmarks_bvult);

use base_ops;
use helper;
use zero_extend;


###################################################
# Unsigned less than
###################################################


sub bvult {
	@_ == 3 || die("ERROR: too few/many arguments for bvult()\n\tusage: bvult(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $bw_next = next_2n($bw);

	local $x_next = declareVar("bvult_x0", $bw_next);
	assert(
		equal(
			$x_next,
			zero_extend($term1, $bw, $bw_next-$bw)
		)
	);
	local $y_next = declareVar("bvult_y0", $bw_next);
	assert(
		equal(
			$y_next,
			zero_extend($term2, $bw, $bw_next-$bw)
		)
	);

	local $x = $x_next;
	local $y = $y_next;
	local $ite_cond;
	
	$bw = $bw_next;

	local $i;
	for ($bw_next/=2, $i=1; $bw_next>=1; $bw_next/=2, $i++) {
		$ite_cond = declareVar("bvult_c$i", 1);
		assert(
			equal(
				$ite_cond,
				equal(
					slice($x, $bw-1, $bw_next),
					slice($y, $bw-1, $bw_next)
				)
			)
		);
		
		$x_next = declareVar("bvult_x$i", $bw_next);
		assert(
			equal(
				$x_next,
				helper::ite(
					$ite_cond,
					slice($x, $bw_next-1, 0),
					slice($x, $bw-1, $bw_next),
					$bw_next
				),
			)
		);
		
		$y_next = declareVar("bvult_y$i", $bw_next);
		assert(
			equal(
				$y_next,
				helper::ite(
					$ite_cond,
					slice($y, $bw_next-1, 0),
					slice($y, $bw-1, $bw_next),
					$bw_next
				)
			)
		);
		
		$x = $x_next;
		$y = $y_next;
		$bw = $bw_next;
	}
	
	return land(lnot($x),$y);
}


sub test_bvult {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvult()\n\tusage: test_bvult(input1,input2,bw)\n");


	# START

	startSmt("bvult");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", 1);
	assert(
		equal(
			$result,
			"(bvult $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvult($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvult {
	@_ == 4 || die("ERROR: too few/many arguments for benchmarks_bvult()\n\tusage: benchmarks_bvult(dir, bw_min, bw_max, bw_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(10))+1; $bw<=2000; $bw+=int(rand(100))) {
		println("Generating benchmark for bvult with bit-width $bw...");
	
		$filename = "$dir/bvult_".sprintf("%04d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvult("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 3 ) {
		print "usage: bvult  dir  bw_min  bw_max  bw_step\n";
		exit;
	}
	benchmarks_bvult($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3]);
}
