#!/usr/bin/perl

package bvurem;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(bvurem benchmarks_bvurem);

use base_ops;
use bvmul;
use bvadd;
use bvult;
use helper;

###################################################
# Unsigned remainder
###################################################


sub bvurem {
	@_ == 3 || die("ERROR: too few/many arguments for bvurem()\n\tusage: bvurem(term1,term2,bw)\n");
  
	local $term1 = $_[0];
	local $term2 = $_[1];
	local $bw = $_[2];

	local $d = declareVar("bvurem_d", $bw);
	local $r = declareVar("bvurem_r", $bw);

	local $bvmul_overflow, $bvadd_overflow;

	assert(
		equal(
			$term1,
			bvadd(
				bvmul(
					$term2,
					$d,
					$bw,
					UNSIGNED,
					$bvmul_overflow
				),
				$r,
				$bw,
				UNSIGNED,
				$bvadd_overflow
	)));

	assert(
		bvult(
			$r,
			$term2,
			$bw
	));

	assert(lnot($bvmul_overflow));
	assert(lnot($bvadd_overflow));

	return $r;
}


sub test_bvurem {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvurem()\n\tusage: test_bvurem(input1,input2,bw)\n");


	# START

	startSmt("bvurem");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	local $input1 = declareVar($input1, $bw);
	local $input2 = declareVar($input2, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvurem $input1 $input2)"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvurem($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvurem {
	@_ == 3 || die("ERROR: too few/many arguments for benchmarks_bvurem()\n\tusage: benchmarks_bvurem(dir, bw_min, bw_max)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw;
	for ($bw = $bw_min; $bw <= $bw_max; $bw++) {
		println("Generating benchmark for bvurem with bit-width $bw...");
	
		$filename = "$dir/bvurem_".sprintf("%02d", $bw).".smt2";
		open FILE, ">", $filename or die $!;
		
		select FILE;
		
		test_bvurem("input1", "input2", $bw);
		
		select STDOUT;
		
		close FILE;
		
#		system("./boolector", "--smt2", "-m", $filename);
	}

}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 2 ) {
		print "usage: bvurem  dir  bw_min  bw_max\n";
		exit;
	}
	benchmarks_bvurem($ARGV[0], $ARGV[1], $ARGV[2]);
}
