#!/usr/bin/perl

package helper;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(selfconcat ite binmagic expand);

use base_ops;
use concat;
use zero_extend;
use bvshl;


sub selfconcat {
	@_ == 3 || die("ERROR: too few/many arguments for selfconcat()\n\tusage: selfconcat(term,bw,bw2)\n");

	local $term = $_[0];
	local $bw = $_[1];
	local $bw2 = $_[2];

	$bw<=$bw2 || die("ERROR when using selfconcat: bw must be less than or equal to bw2\n");

	if ($bw==$bw2) {
		return $term;
	}

	$selfconcat = concat($term, $bw, $term, $bw);

	$bw *= 2;
	while ($bw<$bw2) {
		$selfconcat = concat($selfconcat, $bw, $selfconcat, $bw);
		$bw *= 2;
	}

	return $selfconcat;
}

sub ite {
	@_ == 4 || die("ERROR: too few/many arguments for ite()\n\tusage: ite(cond,term1,term2,bw)\n");

	local $cond = $_[0];
	local $term1 = $_[1];
	local $term2 = $_[2];
	local $bw = $_[3];

	local $cond_term = declareVar("ite_cond", $bw);
	assert(
		equal(
			$cond_term,
			slice(
				selfconcat($cond, 1, $bw),
				$bw-1, 0
			)
	));
	
	return bvor(
		bvand($cond_term, $term1),
		bvand(bvnot($cond_term), $term2)
	);
}

sub binmagic {
	@_ == 2 || die("ERROR: too few/many arguments for binmagic()\n\tusage: binmagic(bw1,bw2)\n");

	local $bw1 = $_[0];
	local $bw2 = $_[1];

	$bw1<$bw2 || die("ERROR when using binmagic: bw1 must be less than bw2\n");
	
	return selfconcat(
		concat(
			bvconst(0, $bw1), $bw1,
			bvnot(bvconst(0, $bw1)), $bw1
		),
		2*$bw1,
		$bw2
	);
}

sub halfshuffle {
	@_ == 3 || die("ERROR: too few/many arguments for halfshuffle()\n\tusage: halfshuffle(term,bw,bw2)\n");

	local $term = $_[0];
	local $bw = $_[1];
	local $bw2 = $_[2];

	$bw<=$bw2 || die("ERROR when using halfshuffle(): bw must not be greater than bw2\n");

	local $m = log2($bw);
	local $n = log2($bw2);

	$m==int($m) && $n==int($n) || die("ERROR when using halfshuffle: bw and bw2 must be powers of 2\n");
	
	local $x, $x_next;
	
	$x = declareVar("halfshuffle_x0", $bw2);
	assert(
		equal(
			$x,
			zero_extend($term, $bw, $bw2-$bw)
	));
	
	local $i;
	for ($i=1; $i<=$m; $i++) {
		$x_next = declareVar("halfshuffle_x$i", $bw2);
		assert(
			equal(
				$x_next,
				bvand(
					bvor(
						$x,
						bvshl_const(
							$x,
							($bw2>>$i) - ($bw>>$i),
							$bw2
						)
					),
					binmagic(
						$bw>>$i,
						$bw2
					)
		)));
		
		$x = $x_next;
	}
	
	return $x;
}

sub expand {
	@_ == 3 || die("ERROR: too few/many arguments for expand()\n\tusage: expand(term,bw,bw2)\n");

	local $term = $_[0];
	local $bw = $_[1];
	local $bw2 = $_[2];

	$bw<=$bw2 || die("ERROR when using expand(): bw must not be greater than bw2\n");

	local $m = log2($bw);
	local $n = log2($bw2);

	$m==int($m) && $n==int($n) || die("ERROR when using expand: bw and bw2 must be powers of 2\n");
	
	local $x, $x_next;
	
	$x = declareVar("expand_x0", $bw2);
	assert(
		equal(
			$x,
			halfshuffle($term, $bw, $bw2)
	));
	
	local $i;
	for ($i=1; $i<=$n-$m; $i++) {
		$x_next = declareVar("expand_x$i", $bw2);
		assert(
			equal(
				$x_next,
				bvor(
					$x,
					bvshl_const(
						$x,
						1<<($i-1),
						$bw2
					)
		)));
		
		$x = $x_next;
	}
	
	return $x;
}


#======= TESTING FUNCTIONS

sub test_expand {
	@_ == 3 || die("ERROR: too few/many arguments for test_expand()\n\tusage: test_expand(input,bw,bw2)\n");


	### SMT HEADER

	printHeader();


	### INPUT VARIABLES

	local $input = $_[0];
	local $bw = $_[1];
	local $bw2 = $_[2];

	local $input = declareVar($input, $bw);


	### CONCRETE VALUES FOR INPUT VARIABLES

	assert(
		equal(
			$input,
			bvconst(37, 8)
	));


	### RESULT BY OUR APPROACH

	local $result = declareVar("result", $bw2);
	assert(
		equal(
			$result,
			expand($input, $bw, $bw2)
	));


	### SMT FOOTER

	printFooter();
}


1;
